{******************************************************************************}
{**                                                                          **}
{**  umEdit Component's set. (TumEdit, TumValidEdit, TumNumberEdit)          **}
{**                                                                          **}
{**  Authors:   Aleksey Kuznetsov and KARPOLAN                               **}
{******************************************************************************}
                    
Unit umEdit;

{$IfDef VER80}  {$Define Delphi3andLower} {$EndIf}
{$IfDef VER90}  {$Define Delphi3andLower} {$EndIf}
{$IfDef VER93}  {$Define Delphi3andLower} {$EndIf}
{$IfDef VER100} {$Define Delphi3andLower} {$EndIf}
{$IfDef VER110} {$Define Delphi3andLower} {$EndIf}

{$IfNdef Delphi3andLower}
  {$Define Delphi4andHigher}
{$EndIf}

Interface

Uses
  {$IfDef Win32} Windows, {$Else Win32} WinTypes, WinProcs, {$EndIf Win32}
  Messages, SysUtils, StdCtrls, Classes, Controls, Graphics, Forms;

Type
{##############################################################################}
{## TumEdit = Class(TEdit) ####################################################}
{##############################################################################}
  TumEdit = Class(TEdit)
  Private
    fFocused   : Boolean;
  {** Messages Routines **}
    Procedure WMPaint(Var Message: TWMPaint); Message WM_PAINT;
    Procedure CMEnter(Var Message: TCMEnter); Message CM_ENTER;
    Procedure CMExit (Var Message: TCMExit ); Message CM_EXIT;

  Protected
    fCanvas    : TControlCanvas;
    fAlignment : TAlignment;
  {** Override this **}
    Procedure Notification(AComponent: TComponent;
                           Operation : TOperation);                    Override;
    Procedure Change;                                                  Override;
  {** Margins Routine **}
    Function  GetTextMargins : TPoint;
  {** Properties Routines **}
    Procedure SetAlignment(A : TAlignment);                             Virtual;
    Procedure SetFocused  (A : Boolean);                                Virtual;
    Function  GetCanvas : TCanvas;                                      Virtual;

  Public
  {** Override this **}
    Constructor Create(AOwner: TComponent);                            Override;
    Destructor  Destroy;                                               Override;
  {** Properties **}
    Property Canvas   : TCanvas
      Read   GetCanvas;

  Published
  {** Properties **}
    Property Alignment    : TAlignment
      Read   fAlignment
      Write  SetAlignment   Default taLeftJustify;
  End;{TumEdit = Class(TEdit)}


{##############################################################################}
{## TumValidEdit = Class(TumEdit) #############################################}
{##############################################################################}
  TumValidEdit = Class(TumEdit)
  Private
    fValidChars    : String;
    fValidateChars : Boolean;
  {** Messages Routines **}
    Procedure WMChar(Var Msg : TWMChar); Message WM_CHAR;
  Public
  {** Override this **}
    Constructor Create(AOwner : TComponent);                           Override;
  Published
  {** Properties **}
    Property ValidChars     : String
      Read   fValidChars
      Write  fValidChars;
    Property ValidateChars  : Boolean
      Read   fValidateChars
      Write  fValidateChars   Default True;
  End;{TumValidEdit = Class(TumEdit)}


{##############################################################################}
{## TumNumberEdit = Class(TumEdit) ############################################}
{##############################################################################}
  TumNumberEditKind = (neDec, neHex, neBin);

  TumNumberEdit = Class(TumEdit)
  Private
    fOnChange   : TNotifyEvent;
  {** Messages Routines **}
    Procedure WMChar(Var Msg : TWMChar); message WM_CHAR;
  Protected
    fKind       : TumNumberEditKind;
    fValue      : LongInt;
    fMaxValue   : LongInt;
    fValidChars : String;
    PrevText    : String;
    PrevValue   : LongInt;
    PrevPos     : Integer;
  {** New OnChanged handler **}
    Procedure NChanged(Sender: TObject);
  {** Properties Routines **}
    Procedure SetValue   (aValue : LongInt);                            Virtual;
    Procedure SetMaxValue(aValue : LongInt);                            Virtual;
    Procedure SetKind    (aValue : TumNumberEditKind);                  Virtual;

  Public
  {** Override this **}
    Constructor Create(AOwner : TComponent);                           Override;

  Published
  {** Properties **}
    Property Kind         : TumNumberEditKind
      Read   fKind
      Write  SetKind        Default neDec;
    Property MaxValue     : LongInt
      Read   fMaxValue
      Write  SetMaxValue;
    Property Value        : LongInt
      Read   fValue
      Write  SetValue;
  {** Events **}
    Property OnChange     : TNotifyEvent
      Read   fOnChange
      Write  FOnChange;
  End;{TumNumberEdit = Class(TumEdit)}


{##############################################################################}
{## Transformations routines ##################################################}
{##############################################################################}
Function DecToHex(aValue : LongInt) : String;
Function HexToDec(aValue : String ) : LongInt;
Function DecToBin(aValue : LongInt) : String;
Function BinToDec(aValue : String ) : LongInt;


Procedure Register;

{##############################################################################}
{******************************************************************************}
{##############################################################################}
Implementation
{##############################################################################}
{******************************************************************************}
{##############################################################################}
{$R *.RES}

Procedure Register;
Begin
  RegisterComponents('UtilMind', [TumEdit, TumValidEdit, TumNumberEdit]);
End;{Procedure Register}


{##############################################################################}
{## Transformations routines ##################################################}
{##############################################################################}
Function DecToHex(aValue : LongInt) : String;
Var
  w  : Array[1..2] of Word absolute aValue;
  St : String;

  Function HexByte(b: Byte): String;
  Const
    Hex: Array[$0..$F] of Char = '0123456789ABCDEF';
  Begin
    HexByte := Hex[b shr 4] + Hex[b and $F];
  End;

  Function HexWord(w: Word): String;
  Begin
    HexWord := HexByte(Hi(w)) + HexByte(Lo(w));
  End;

Begin
  St := HexWord(w[2]) + HexWord(w[1]);
  while (St[1] = '0') and (Length(St) > 1) do
   Delete(St, 1, 1);
  Result := St;
End;{Function DecToHex}

Function HexToDec(aValue : String) : LongInt;
Var
  l : LongInt;
  b : Byte;
Begin
  Result := 0;
  if Length(aValue) <> 0 then
   begin
     l := 1;
     b := Length(aValue) + 1;
     repeat
      dec(b);
      if aValue[b] <= '9' then Result := Result + (Byte(aValue[b]) - 48) * l
      else Result := Result + (Byte(aValue[b]) - 55) * l;
      l := l * 16;
     until b = 1;
   end;{if Length(aValue) <> 0 then}
End;{Function HexToDec}

Function DecToBin(aValue : LongInt) : String;
Var
  w  : Array[1..2] of Word absolute aValue;
  St : String;

  Function BinByte(b: Byte): String;
  Const
    Bin: Array[False..True] of Char = '01';
  Begin
    Result := Bin[b and 128 = 128] + Bin[b and 64 = 64] + Bin[b and 32 = 32] + Bin[b and 16 = 16] +
              Bin[b and 8 = 8] + Bin[b and 4 = 4] + Bin[b and 2 = 2] + Bin[b and 1 = 1];
  End;

  Function BinWord(w: Word) : String;
  Begin
    BinWord := BinByte(Hi(w)) + BinByte(Lo(w));
  End;

Begin
  St := BinWord(w[2]) + BinWord(w[1]);
  while (St[1] = '0') and (Length(St) > 1) do
   Delete(St, 1, 1);
  Result := St;
End;{Function DecToBin}

Function BinToDec(aValue : String) : LongInt;
Var
  l : LongInt;
  b : Byte;
Begin
  Result := 0;
  if Length(aValue) = 0
   then Exit;

  l := 1;
  b := Length(aValue) + 1;
  repeat
   dec(b);
   if aValue[b] = '1'
    then Result := Result + l;
   l := l shl 1;
  until b = 1;
End;{Function BinToDec}



{##############################################################################}
{## TumEdit = Class(TEdit) ####################################################}
{##############################################################################}
{******************************************************************************}
{** Public ********************************************************************}
{******************************************************************************}
{** Override this **}
{*******************}
Constructor TumEdit.Create(AOwner: TComponent);
Begin
  fAlignment := taLeftJustify;

  Inherited Create(AOwner);
  fCanvas := TControlCanvas.Create;
  fCanvas.Control := Self;
End;{Constructor TumEdit.Create}

Destructor TumEdit.Destroy;
Begin
  fCanvas.Free;
  Inherited Destroy;
End;{Destructor TumEdit.Destroy}


{******************************************************************************}
{** Protected *****************************************************************}
{******************************************************************************}
{** Override this **}
{*******************}
Procedure TumEdit.Notification(AComponent: TComponent;
                               Operation : TOperation);
Begin
  Inherited Notification(AComponent, Operation);
End;{Procedure TantTaskbarIcon.Notification}

Procedure TumEdit.Change;
Begin
  Inherited Change;
{  Invalidate; {!!! WARNING Bug on loosing focus (KARPOLAN)}
End;{Procedure TumEdit.Change}


{*********************}
{** Margins Routine **}
{*********************}
Function TumEdit.GetTextMargins : TPoint;
Var
  DC       : HDC;
  SaveFont : HFont;
  i        : Integer;
  SysMetrics, Metrics : TTextMetric;
Begin
  if NewStyleControls then
   begin
     if BorderStyle = bsNone
      then i := 0
      else
     if Ctl3D
      then i := 1
      else i := 2;
     Result.X := SendMessage(Handle, EM_GETMARGINS, 0, 0) and $0000FFFF + i;
     Result.Y := i;
   end else
   begin
     if BorderStyle = bsNone
      then i := 0
      else
        begin
          DC := GetDC(0);
          GetTextMetrics(DC, SysMetrics);
          SaveFont := SelectObject(DC, Font.Handle);
          GetTextMetrics(DC, Metrics);
          SelectObject(DC, SaveFont);
          ReleaseDC(0, DC);
          i := SysMetrics.tmHeight;
          if i > Metrics.tmHeight
           then i := Metrics.tmHeight;
          i := i div 4;
        end;{if BorderStyle else}
     Result.X := i;
     Result.Y := i;
   end;{if NewStyleControls else}
End;{Function TumEdit.GetTextMargins}


{***********************}
{** Propertie Routine **}
{***********************}
Procedure TumEdit.SetAlignment(A : TAlignment);
Begin
  if fAlignment = A
   then Exit;
  fAlignment := A;
  Invalidate;
End;{Procedure TumEdit.SetAlignment}

Procedure TumEdit.SetFocused(A : Boolean);
Begin
  if fFocused = A
   then Exit;
  fFocused := A;
{  Invalidate; {!!! WARNING Bug on loosing focus (KARPOLAN)}
End;{Procedure TumEdit.SetFocused}

Function TumEdit.GetCanvas : TCanvas;
Begin
  Result := TCanvas(fCanvas);
End;{Function TumEdit.GetCanvas}


{******************************************************************************}
{** Private *******************************************************************}
{******************************************************************************}
{** Message Routines **}
{**********************}
Procedure TumEdit.WMPaint(Var Message : TWMPaint);
{$IfDef Delphi4andHigher}
Const
  AlignStyle : Array[Boolean, TAlignment] Of DWord =
   ((WS_EX_LEFT , WS_EX_RIGHT, WS_EX_LEFT),
    (WS_EX_RIGHT, WS_EX_LEFT , WS_EX_LEFT));
{$EndIf Delphi4andHigher}
Var
  ALeft       : integer;
  Margins     : TPoint;
  R           : TRect;
  DC          : HDC;
  PS          : TPaintStruct;
  strText     : String;
  AAlignment  : TAlignment;
{$IfDef Delphi4andHigher}
  ExStyle     : DWord;
{$EndIf Delphi4andHigher}

  Procedure TryToPaint;
  Begin
    fCanvas.Font := Font;
    with fCanvas do
     begin
       R := ClientRect;
       if not (NewStyleControls and Ctl3D) and (BorderStyle = bsSingle) then
        begin
          Brush.Color := clWindowFrame;
          FrameRect(R);
          InflateRect(R, -1, -1);
        end;{if not (NewSt... }
       Brush.Color := Color;
       if not Enabled
        then Font.Color := clGrayText;
       strText := Text;
       if (csPaintCopy in ControlState) then
        begin
          case CharCase of
           ecUpperCase: strText := AnsiUpperCase(strText);
           ecLowerCase: strText := AnsiLowerCase(strText);
          end;
        end;{if (csPaintCopy in ControlState) then}
       if PasswordChar <> #0
        then FillChar(strText[1], Length(strText), PasswordChar);
       Margins := GetTextMargins;
       case AAlignment of
        taLeftJustify  : ALeft := Margins.X;
        taRightJustify : ALeft := ClientWidth - TextWidth(strText) - Margins.X - 1;
        else             ALeft := (ClientWidth - TextWidth(strText)) div 2;
       end;{case AAlignment of}

{$IfDef Delphi4andHigher}
       if SysLocale.MiddleEast
        then UpdateTextFlags;
{$EndIf Delphi4andHigher}

       TextRect(R, ALeft, Margins.Y, strText);
     end;{with fCanvas do}
  End;{INTERNAL Procedure TryToPaint}

  Procedure PaintDefault;
  Begin
  {** Prepare Canvas **}
    DC := Message.DC;
    if DC = 0
     then DC := BeginPaint(Handle, PS);
    fCanvas.Handle := DC;
  {** Try Paint **}
    try
      TryToPaint;
    finally
      fCanvas.Handle := 0;
      if Message.DC = 0
       then EndPaint(Handle, PS);
    end;{try..finally}
  End;{INTERNAL Procedure PaintDefault}

Begin
{** Alignment depend Focused state **}
  if fFocused
   then AAlignment := taLeftJustify
   else AAlignment := fAlignment;

{$IfDef Delphi4andHigher}
{** Update RightToLeftAlignment **}
  if UseRightToLeftAlignment
   then ChangeBiDiModeAlignment(AAlignment);
{$EndIf Delphi4andHigher}

{** Default Drawing if Focused **}
  if ((AAlignment = taLeftJustify) or FFocused) and
     (not (csPaintCopy in ControlState)) then
   begin

{$IfDef Delphi4andHigher}
     if SysLocale.MiddleEast and HandleAllocated and (IsRightToLeft) then
      begin { This keeps the right aligned text, right aligned }
        ExStyle := DWORD(GetWindowLong(Handle, GWL_EXSTYLE)) and (not WS_EX_RIGHT) and
          (not WS_EX_RTLReadING) and (not WS_EX_LEFTSCROLLBAR);
        if UseRightToLeftReading then ExStyle := ExStyle or WS_EX_RTLReadING;
        if UseRightToLeftScrollbar then ExStyle := ExStyle or WS_EX_LEFTSCROLLBAR;
        ExStyle := ExStyle or
          AlignStyle[UseRightToLeftAlignment, AAlignment];
        if DWORD(GetWindowLong(Handle, GWL_EXSTYLE)) <> ExStyle then
          SetWindowLong(Handle, GWL_EXSTYLE, ExStyle);
      end;{if SysLocale.MiddleEast ...}
{$EndIf Delphi4andHigher}

     Inherited;
     Exit;
   end;{if ((AAlignment = taLeftJustify)...}

  PaintDefault;
  Inherited;
End;{Procedure TumEdit.WMPaint}

Procedure TumEdit.CMEnter(Var Message : TCMEnter);
Begin
  SetFocused(True);
  Inherited;
End;{Procedure TumEdit.CMEnter}

Procedure TumEdit.CMExit(Var Message : TCMExit);
Begin
  Inherited;
  SetFocused(False);
  Invalidate;
End;{Procedure TumEdit.CMExit}


{##############################################################################}
{## TumValidEdit = Class(TumEdit) #############################################}
{##############################################################################}
{******************************************************************************}
{** Public ********************************************************************}
{******************************************************************************}
{** Override this **}
{*******************}
Constructor TumValidEdit.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  fValidateChars := True;
End;{Constructor TumValidEdit.Create}

{******************************************************************************}
{** Private *******************************************************************}
{******************************************************************************}
{** Messages Routines **}
{***********************}
Procedure TumValidEdit.WMChar(Var Msg : TWMChar);
Var
  i : Integer;
Begin
  if ValidateChars and (Char(Msg.CharCode) <> #8) then { Enable BackSpace key }
   begin
    i := Length(FValidChars);
    if i <> 0 then
     for i := 1 to i do
      if fValidChars[i] = Char(Msg.CharCode)
       then Inherited
   end
  else Inherited;
End;{Procedure TumValidEdit.WMChar}


{##############################################################################}
{## TumNumberEdit = Class(TumEdit) ############################################}
{##############################################################################}
{******************************************************************************}
{** Public ********************************************************************}
{******************************************************************************}
{** Override this **}
{*******************}
Constructor TumNumberEdit.Create(AOwner : TComponent);
Begin
  SetKind(neDec);
  Inherited Create(AOwner);
  Text := '0';
  Inherited OnChange := NChanged;
End;{Constructor TumNumberEdit.Create}


{******************************************************************************}
{** Protected *****************************************************************}
{******************************************************************************}
{** New OnChanged handler **}
{***************************}
Procedure TumNumberEdit.NChanged(Sender : TObject);
Label ex;
Var
  l     : LongInt;
  St    : String;
  Error : Boolean;
Begin
  if (csDesigning in ComponentState)
   then Exit;

  if (csLoading in ComponentState) or (csReading in ComponentState)
   then Exit;

  if Text = '' then
   begin
     FValue := 0;
     goto ex;
   end;{if Text = '' then}

  l := 0;

  if (FKind = neDec) then
   begin
     Error := False;
     try
       l := StrToInt(Text)
     except
       Error := True;
       Text := PrevText;
       SelStart := PrevPos;
       SelLength := 0;
     end;{try..except}
     if Error then goto ex;
   end else {if (FKind = neDec)}
  if (FKind = neHex)
   then l := HexToDec(Text)
   else l := BinToDec(Text);

  if (FMaxValue <> 0) then
   begin
     if FKind = neDec
      then St := IntToStr(FMaxValue)
      else
     if FKind = neHex
      then St := DecToHex(FMaxValue)
      else St := DecToBin(FMaxValue);

     if (l > FMaxValue) or (l < 0) or (Length(Text) > Length(St)) or
        ((Length(Text) = Length(St)) and (Text > St)) then
      begin
        Text := PrevText;
        SelStart := PrevPos;
        SelLength := 0;
        goto ex;
      end;{if (l > FMaxValue) or ...}
   end;{if (FMaxValue <> 0) then}

  FValue := l;
  PrevText := Text;
  PrevValue := FValue;

ex: {Label}
  if Assigned(FOnChange)
   then FOnChange(Sender);
End;{Procedure TumNumberEdit.NChanged}


{*************************}
{** Properties Routines **}
{*************************}
Procedure TumNumberEdit.SetValue(aValue : LongInt);
Begin
  if (aValue < 0)
   then aValue := 0
   else
  if (FMaxValue <> 0) and (aValue > FMaxValue)
   then aValue := FMaxValue;

  FValue := aValue;
  if (Kind = neHex)
   then Text := DecToHex(aValue)
   else
  if (Kind = neDec)
   then Text := IntToStr(aValue)
   else Text := DecToBin(aValue)
End;{Procedure TumNumberEdit.SetValue}

Procedure TumNumberEdit.SetMaxValue(aValue : LongInt);
Begin
  FMaxValue := aValue;
  if (FMaxValue <> 0) and (FValue > FMaxValue)
   then Value := FMaxValue;
End;{Procedure TumNumberEdit.SetMaxValue}

Procedure TumNumberEdit.SetKind(aValue : TumNumberEditKind);
Begin
  FKind := aValue;
  if aValue = neDec
   then FValidChars := '0123456789'
   else
  if aValue = neHex
   then FValidChars := '0123456789abcdefABCDEF'
   else
  if aValue = neBin
   then FValidChars := '01'
   else FValidChars := '01234567';  {** Octacimal **}
End;{Procedure TumNumberEdit.SetKind}


{******************************************************************************}
{** Private *******************************************************************}
{******************************************************************************}
{** Messages Routines **}
{***********************}
Procedure TumNumberEdit.WMChar(Var Msg : TWMChar);
Var
  i : Integer;
Begin
  PrevText  := Text;
  PrevValue := FValue;
  PrevPos   := SelStart;
  if Char(Msg.CharCode) = #8 then Inherited
  else
   begin
     i := Length(FValidChars);
     if i <> 0 then
      for i := 1 to i do
       if FValidChars[i] = Char(Msg.CharCode) then
        Inherited
   end;
End;{Procedure TumNumberEdit.WMChar}


END{Unit umEdit}.
